#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <SoftwareServo.h>
#include <avr/power.h>

const byte servoPin = 3;  // transistor, servo power
const byte ledPin = 4;    // LED to indicate when the microcontroller is awake
byte keep_ADCSRA;         // what ADC items are enabled normally
const byte gearLow = 1;
const byte gearHigh = 6;
byte gearPos = 1;      // Current gear position
byte gearPosAddr = 0;  // store here
byte servoPos = 0;     // actual servo position value associated with gear up or down shift
volatile bool shiftUpRequested = false;
volatile bool shiftDownRequested = false;
volatile bool awake = false;

// Timing variable for cooldown
unsigned long lastShiftTime = 0;
const unsigned long cooldownDelay = 10; // Cooldown after shift

SoftwareServo myservo;

void setup() {
  cli();  // Disable interrupts
  keep_ADCSRA = ADCSRA;
  ADCSRA = 0;                   // disable ADC to save power during sleep
  myservo.attach(2);            // attaches the servo on pin 2 to the servo object
  pinMode(servoPin, OUTPUT);    // Power for Servo
  digitalWrite(servoPin, LOW);  // Start with the servo off
  pinMode(ledPin, OUTPUT);
  digitalWrite(ledPin, LOW);    // Start with the LED off

  // Set up pin change interrupt on PB0 and PB1
  pinMode(0, INPUT_PULLUP);                // Pin PB0 (Shift Up)
  pinMode(1, INPUT_PULLUP);                // Pin PB1 (Shift Down)
  GIMSK |= (1 << PCIE);                    // Enable pin change interrupts
  PCMSK |= (1 << PCINT0) | (1 << PCINT1);  // Enable interrupts on PB0 & PB1

  sei();  // Enable interrupts
}

// Interrupt Service Routine for pin change interrupts
ISR(PCINT0_vect) {
  awake = true;  // Set the awake flag when an interrupt occurs
  if (digitalRead(0) == LOW) {
    shiftUpRequested = true;   // Mark the Shift Up request
  }
  if (digitalRead(1) == LOW) {
    shiftDownRequested = true; // Mark the Shift Down request
  }
}

void loop() {
  if (awake) {
    // Microcontroller has woken up
    ADCSRA = keep_ADCSRA;
    digitalWrite(servoPin, HIGH);  // Turn on servo, we're awake
    digitalWrite(ledPin, HIGH);    // Turn on the LED to indicate we're awake

    unsigned long currentTime = millis();

    // Handle Shift Up with cooldown
    if (shiftUpRequested && (currentTime - lastShiftTime) > cooldownDelay) {
      shiftUpRequested = false;  // Reset flag
      sUP();
      lastShiftTime = currentTime;  // Update cooldown timer
    }

    // Handle Shift Down with cooldown
    if (shiftDownRequested && (currentTime - lastShiftTime) > cooldownDelay) {
      shiftDownRequested = false;  // Reset flag
      sDOWN();
      lastShiftTime = currentTime;  // Update cooldown timer
    }

    // Turn off the servo after the operation
    SoftwareServo::refresh();
    delay(100);
  }

  // Prepare for sleep
  cli();          // Disable interrupts
  awake = false;  // Reset the awake flag
  keep_ADCSRA = ADCSRA;
  ADCSRA = 0;  // disable ADC
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  sleep_enable();
  sei();        // Re-enable interrupts
  sleep_cpu();  // Enter sleep mode

  // Code resumes here after waking up
  sleep_disable();  // Disable sleep mode
}

void sUP() {
  if (gearPos < gearHigh) {
    gearPos++;
  }

  switch (gearPos) {
    case 2:
      servoPos = 34;
      break;
    case 3:
      servoPos = 43;
      break;
    case 4:
      servoPos = 51;
      break;
    case 5:
      servoPos = 59;
      break;
    case 6:
      servoPos = 73;
      break;
  }
  save();
}

void sDOWN() {
  if (gearPos > gearLow) {
    gearPos--;
  }

  switch (gearPos) {
    case 5:
      servoPos = 42;
      break;
    case 4:
      servoPos = 33;
      break;
    case 3:
      servoPos = 26;
      break;
    case 2:
      servoPos = 19;
      break;
    case 1:
      servoPos = 11;
      break;
  }
  save();
}

void save() {
  for (int i = 0; i < 4; i++) {
    myservo.write(servoPos);  
    delay(10);
    SoftwareServo::refresh();  
    delay(50);                 
  }

  digitalWrite(servoPin, LOW);
  digitalWrite(ledPin, LOW);
}
